import os
from typing import Union
import requests
import random
from .get_kachery_cloud_dir import get_kachery_cloud_dir
from ._kacherycloud_request import _kacherycloud_request


def load_file(uri: str, *, verbose: bool=False) -> Union[str, None]:
    assert uri.startswith('ipfs://'), f'Invalid or unsupported URI: {uri}'
    a = uri.split('?')[0].split('/')
    assert len(a) >= 3, f'Invalid or unsupported URI: {uri}'
    cid = a[2]

    kachery_cloud_dir = get_kachery_cloud_dir()
    e = cid[-6:]
    parent_dir = f'{kachery_cloud_dir}/ipfs/{e[0]}{e[1]}/{e[2]}{e[3]}/{e[4]}{e[5]}'
    filename = f'{parent_dir}/{cid}'
    if os.path.exists(filename):
        return filename

    tmp_filename = filename + '.downloading'
    if os.path.exists(tmp_filename):
        raise Exception(f'Temporary file exists.')

    payload = {
        'type': 'findIpfsFile',
        'cid': cid
    }
    response= _kacherycloud_request(payload)
    found = response['found']
    if found:
        url = response['url']
    else:
        raise Exception(f'File not found: {uri}')
        # url = f'https://{cid}.ipfs.dweb.link'
        # url = f'https://cloudflare-ipfs.com/ipfs/{cid}'
        # url = f'https://ipfs.filebase.io/ipfs/{cid}'

    if verbose:
        print(f'Loading file from kachery cloud: {uri}')    
    if not os.path.exists(parent_dir):
        os.makedirs(parent_dir)
    tmp_filename = f'{filename}.tmp.{_random_string(8)}'
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(tmp_filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)
    try:
        os.rename(tmp_filename, filename)
    except:
        if not os.path.exists(filename): # maybe some other process beat us to it
            raise Exception(f'Unexpected problem moving file {tmp_filename}')
    return filename

def _random_string(num_chars: int) -> str:
    chars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
    return ''.join(random.choice(chars) for _ in range(num_chars))


import time
from typing import List, Union

from kachery_cloud.TaskBackend.PubsubListener import PubsubListener
from kachery_cloud.load_file import _random_string
from .._kacherycloud_request import _kacherycloud_request
from ..get_client_id import get_client_id


class Feed:
    def __init__(self, *, feed_id: str, project_id: str) -> None:
        self._feed_id = feed_id
        self._project_id = project_id
        self._current_message_number = 0
        self._pubsub_listener: Union[PubsubListener, None] = None
        self._project_id = project_id
        self._messages_appended_callbacks = {}
    def append_messages(self, messages: List[dict]):
        payload = {
            'type': 'appendFeedMessages',
            'feedId': self._feed_id,
            'messages': messages
        }
        resp = _kacherycloud_request(payload)
    @property
    def uri(self):
        return f'kachery-feed://{self._feed_id}'
    @property
    def feed_id(self):
        return self._feed_id
    @property
    def current_message_number(self):
        return self._current_message_number
    def get_next_messages(self, *, timeout_sec: float=0):
        payload = {
            'type': 'getFeedMessages',
            'feedId': self._feed_id,
            'startMessageNumber': self._current_message_number
        }
        resp = _kacherycloud_request(payload)
        messages = resp['messages']
        if len(messages) == 0 and timeout_sec > 0:
            return self._wait_for_next_messages(timeout_sec=timeout_sec)
        start_message_number = resp['startMessageNumber']
        self._current_message_number = start_message_number + len(messages)
        return messages
    @staticmethod
    def create(*, project_id: Union[str, None]=None):
        payload = {
            'type': 'createFeed'
        }
        if project_id is not None:
            payload['projectId'] = project_id
        response = _kacherycloud_request(payload)
        feed_id = response['feedId']
        project_id = response['projectId']
        return Feed(feed_id=feed_id, project_id=project_id)
    @staticmethod
    def load(feed_id: str):
        payload = {
            'type': 'getFeedInfo',
            'feedId': feed_id
        }
        response = _kacherycloud_request(payload)
        project_id = response['projectId']
        return Feed(feed_id=feed_id, project_id=project_id)
    def _wait_for_next_messages(self, *, timeout_sec: float):
        timer = time.time()
        if self._pubsub_listener is None:
            self._initialize_pubsub_listener()
        callback_id = _random_string(10)
        got_new_messages = False
        def on_new_messages():
            nonlocal got_new_messages
            got_new_messages = True
        self._messages_appended_callbacks[callback_id] = on_new_messages
        try:
            messages0 = self.get_next_messages()
            if len(messages0) > 0:
                return messages0
            while True:
                if got_new_messages:
                    messages1 = self.get_next_messages()
                    if len(messages1) > 0:
                        return messages1
                elapsed = time.time() - timer
                if elapsed > timeout_sec:
                    return []
                self._pubsub_listener.wait(0.05)
        finally:
            del self._messages_appended_callbacks[callback_id]
        
    def _initialize_pubsub_listener(self):
        client_id = get_client_id()

        def subscribe_to_channel():
            payload_sub = {
                'type': 'subscribeToPubsubChannel',
                'channelName': 'feedUpdates',
                'projectId': self._project_id
            }
            response_sub = _kacherycloud_request(payload_sub)
            return response_sub
        
        response_sub = subscribe_to_channel()
        subscribe_key = response_sub['subscribeKey']
        subscribe_token = response_sub['token']
        pubsub_channel_name = response_sub['pubsubChannelName']

        def handle_message(*, channel: str, message: dict):
            if message['type'] == 'feedMessagesAppended':
                if message['feedId'] == self._feed_id:
                    callbacks = self._messages_appended_callbacks.values()
                    for cb in callbacks:
                        cb()
        def renew_access_token():
            r = subscribe_to_channel()
            return r['token']
        listener = PubsubListener(
            channels=[pubsub_channel_name],
            uuid=client_id,
            subscribe_key=subscribe_key,
            access_token=subscribe_token,
            renew_access_token_callback=renew_access_token
        )
        listener.on_message(handle_message)
        listener.start()
        self._pubsub_listener = listener


import requests
from datetime import datetime
from .get_client_id import get_client_id
from ._client_keys import _sign_message_as_client
from ._kachery_cloud_api_url import _kachery_cloud_api_url

def _kacherycloud_request(request_payload: dict):
    client_id = get_client_id()
    url = f'{_kachery_cloud_api_url}/api/kacherycloud'
    timestamp = int(datetime.timestamp(datetime.now()) * 1000)
    payload = {**request_payload, **{'timestamp': timestamp}}
    req = {
        'payload': payload,
        'fromClientId': client_id,
        'signature': _sign_message_as_client(payload)
    }
    resp = requests.post(url, json=req)
    if resp.status_code != 200:
        raise Exception(f'Error in {payload["type"]} ({resp.status_code}) {resp.reason}: {resp.text}')
    response = resp.json()
    return response


from ._kacherycloud_request import _kacherycloud_request

_global_data = {
    'project_base_urls': {}
}

def _get_project_bucket_base_url(project_id: str):
    if project_id in _global_data['project_base_urls']:
        return _global_data['project_base_urls'][project_id]
    payload = {
        'type': 'getProjectBucketBaseUrl',
        'projectId': project_id
    }
    response = _kacherycloud_request(payload)
    found = response['found']
    if not found:
        raise Exception(f'Project not found: {project_id}')
    bucket_base_url = response['bucketBaseUrl']
    if not bucket_base_url:
        raise Exception('Unexpected, no bucket base url')
    _global_data['project_base_urls'][project_id] = bucket_base_url
    return bucket_base_url


import socket
import urllib.parse

from .get_client_id import get_client_id
from ._client_keys import _sign_message_as_client
from ._kachery_cloud_api_url import _kachery_cloud_api_url
from ._kacherycloud_request import _kacherycloud_request


_global_init = {
    'client_info': 0 # 0 means we have not yet queried. None means not found.
}

def _get_client_info():
    if _global_init['client_info'] != 0:
        return _global_init['client_info']
    client_id = get_client_id()
    payload = {
        'type': 'getClientInfo',
        'clientId': client_id
    }
    response = _kacherycloud_request(payload)
    found = response['found']
    if found:
        _global_init['client_info'] = response
        return response
    else:
        return None

def init():
    client_info = _get_client_info()
    if client_info is None:
        client_id = get_client_id()
        signature = _sign_message_as_client({
            'type': 'addClient'
        })
        label = socket.gethostname()
        url = f'{_kachery_cloud_api_url}/registerClient/{client_id}?signature={signature}&label={urllib.parse.quote(label)}'
        print('')
        print(url)
        print('')
        print('Click the above link to register this kachery cloud client. After pressing the "Yes, proceed" button on the website, press [Enter] in this terminal to continue.')
        print('')
        input()
        print('Checking...')
        client_info = _get_client_info()
        if client_info is None:
            raise Exception('Failed to initialize this kachery client')
        print(f'Client initialized successfully.')
    else:
        print('This client has already been registered.')
    client = client_info['client']
    label = client['label']
    default_project_id = client.get('defaultProjectId', None)
    client_owner = client['ownerId']
    print(f'Label: {label}')
    print(f'Owner: {client_owner}')
    print(f'Default project ID: {default_project_id}')


import click
import kachery_cloud as kc


@click.group(help="Kachery cloud command-line client")
def cli():
    pass

@click.command(help="One-time initialization of the kachery cloud client")
def init():
    kc.init()

@click.command(help="Store file in IPFS")
@click.argument('filename')
def store_file(filename: str):
    uri = kc.store_file(filename)
    print(uri)

@click.command(help="Load file from IPFS")
@click.argument('uri')
def load_file(uri: str):
    fname = kc.load_file(uri)
    print(fname)

@click.command(help="Load file from IPFS and write to stdout")
@click.argument('uri')
def cat_file(uri: str):
    kc.cat_file(uri)

cli.add_command(init)
cli.add_command(store_file)
cli.add_command(load_file)
cli.add_command(cat_file)



import sys
import os
from .load_file import load_file

def cat_file(uri: str):
    old_stdout = sys.stdout
    sys.stdout = None

    local_fname = load_file(uri)

    sys.stdout = old_stdout

    with open(local_fname, 'rb') as f:
        while True:
            data = os.read(f.fileno(), 4096)
            if len(data) == 0:
                break
            os.write(sys.stdout.fileno(), data)



import string
import random
import time
from typing import Callable, Dict, Union
from .TaskClient import TaskClient
from ._run_task import _run_task

try:
    from dask.distributed import Client, Future, LocalCluster, TimeoutError
except:
    raise Exception('Unable to import from dask.distributed. Use: pip install dask[distributed]')


class TaskBackend:
    def __init__(self, *, project_id: Union[str, None]=None, num_workers=1, threads_per_worker=4) -> None:
        self._project_id = project_id
        self._task_client = TaskClient(project_id=project_id)
        self._registered_task_handlers: Dict[str, TaskHandler] = {}
        self._dask_cluster = LocalCluster(n_workers=num_workers, threads_per_worker=threads_per_worker)
        self._dask_client = Client(self._dask_cluster)
        self._task_jobs: Dict[str, TaskJob] = {}
    def register_task_handler(self, *, task_type: str, task_name: str, task_function: Callable):
        self._registered_task_handlers[task_name] = TaskHandler(
            task_type=task_type,
            task_name=task_name,
            task_function=task_function
        )
    def run(self):
        def handle_task_request(*, task_type: str, task_name: str, task_input: dict, task_job_id: str):
            if task_name in self._registered_task_handlers:
                task_handler = self._registered_task_handlers[task_name]
                if task_handler._task_type != task_type:
                    raise Exception(f'Mismatch in task type: {task_handler.task_type} <> {task_type}')
                if task_job_id in self._task_jobs:
                    tj = self._task_jobs[task_job_id]
                    try:
                        result_future: Future = tj._result_future
                        del self._task_jobs[task_job_id]
                    except TimeoutError:
                        # still running, it seems
                        return
                    except Exception:
                        raise
                result_future: Future = self._dask_client.submit(
                    _run_task,
                    pure=False,
                    task_type=task_handler._task_type,
                    task_name=task_handler._task_name,
                    task_job_id=task_job_id,
                    task_function=task_handler._task_function,
                    task_input=task_input,
                    project_id=self._project_id
                )
                # seems to be important to store the result future in memory
                # hypothesis: if the result is not used, it may get garbage collected and not actually run
                self._task_jobs[task_job_id] = TaskJob(
                    task_type=task_handler._task_type,
                    task_name=task_handler._task_name,
                    task_input=task_input,
                    task_job_id=task_job_id,
                    result_future=result_future
                )
        listener = self._task_client.listen_for_task_requests(handle_task_request)
        try:
            while True:
                listener.wait(1)
        finally:
            listener.stop()

class TaskHandler:
    def __init__(self, *, task_type: str, task_name: str, task_function: Callable) -> None:
        self._task_type = task_type
        self._task_name = task_name
        self._task_function = task_function
    def run_task(self, *, task_input: dict):
        return self._task_function(**task_input)

class TaskJob:
    def __init__(self, *,
        task_type: str,
        task_name: str,
        task_input: dict,
        task_job_id: str,
        result_future: Future
    ) -> None:
        self._task_type = task_type
        self._task_name = task_name
        self._task_input = task_input
        self._task_job_id = task_job_id
        self._result_future = result_future

def _random_string(num: int):
    letters = string.ascii_letters
    return ''.join(random.choice(letters) for i in range(num))


from typing import Any, Union, Protocol
import random
import time

from .upload_task_result import upload_task_result, download_task_result
from .PubsubListener import PubsubListener
from .._kacherycloud_request import _kacherycloud_request
from ..get_client_id import get_client_id
from .._json_stringify_deterministic import _json_stringify_deterministic
from .._sha1_of_string import _sha1_of_string

class TaskRequestCallback(Protocol):
    def __call__(self, *, task_type: str, task_name: str, task_input: dict, task_job_id: str): ...

class TaskClient:
    def __init__(self, *, project_id: Union[None, str]=None) -> None:
        self._project_id = project_id
    def download_task_result(self, *, task_name: str, task_job_id: str):
        return download_task_result(task_name=task_name, task_job_id=task_job_id)
    def request_task(self, *, task_type: str, task_name: str, task_input: dict):
        client_id = get_client_id()

        payload_sub = {
            'type': 'subscribeToPubsubChannel',
            'channelName': 'provideTasks'
        }
        if self._project_id is not None:
            payload_sub['projectId'] = self._project_id
        response_sub = _kacherycloud_request(payload_sub)
        subscribe_key = response_sub['subscribeKey']
        subscribe_token = response_sub['token']
        pubsub_channel_name = response_sub['pubsubChannelName']

        status: Union[str, None] = None
        result: Union[Any, None] = None
        error: Union[str, None] = None

        def handle_message(*, channel: str, message: dict):
            nonlocal status
            nonlocal error
            nonlocal result
            if message['type'] == 'setTaskStatus':
                if message['taskName'] == task_name:
                    if message['taskJobId'] == task_job_id:
                        if status == 'error':
                            error = message['error']
                        elif status == 'finished':
                            result = self.download_task_result(task_name=task_name, task_job_id=task_job_id)
                        status = message['status']
        def renew_access_token():
            raise Exception('not used')
        listener = PubsubListener(
            channels=[pubsub_channel_name],
            uuid=client_id,
            subscribe_key=subscribe_key,
            access_token=subscribe_token,
            renew_access_token_callback=renew_access_token
        )
        listener.on_message(handle_message)
        listener.start()
        
        if task_type == 'calculation':
            task_job_id = _sha1_of_string(_json_stringify_deterministic({
                'taskName': task_name,
                'taskInput': task_input
            }))
        else:
            task_job_id = _sha1_of_string(_random_string(100))
        payload = {
            'type': 'publishToPubsubChannel',
            'channelName': 'requestTasks',
            'message': {
                'type': 'requestTask',
                'taskType': task_type,
                'taskName': task_name,
                'taskInput': task_input,
                'taskJobId': task_job_id
            }
        }
        if self._project_id is not None:
            payload['projectId'] = self._project_id
        _kacherycloud_request(payload)
        
        announced_started = False
        try:
            while True:
                listener.wait(0.1)
                if status == 'finished':
                    return result
                elif status == 'error':
                    raise Exception(error)
                elif status == 'started':
                    if not announced_started:
                        print('Task started')
                        announced_started = True
        finally:
            listener.stop()

    def listen_for_task_requests(self, callback: TaskRequestCallback):
        def subscribe_to_request_tasks_channel():
            payload = {
                'type': 'subscribeToPubsubChannel',
                'channelName': 'requestTasks'
            }
            if self._project_id is not None:
                payload['projectId'] = self._project_id
            return _kacherycloud_request(payload)
        def renew_access_token():
            r = subscribe_to_request_tasks_channel()
            return r['token']
        def handle_message(*, channel: str, message: dict):
            if message['type'] == 'requestTask':
                task_type = message['taskType']
                task_name = message['taskName']
                task_input = message['taskInput']
                task_job_id = message['taskJobId']
                if task_type == 'calculation':
                    check_task_job_id = _sha1_of_string(_json_stringify_deterministic({
                        'taskName': task_name,
                        'taskInput': task_input
                    }))
                    if check_task_job_id != task_job_id:
                        raise Exception('Mismatch between task job ID and computed task job ID')
                callback(task_type=task_type, task_name=task_name, task_input=task_input, task_job_id=task_job_id)
                
        client_id = get_client_id()
        response = subscribe_to_request_tasks_channel()
        listener = PubsubListener(
            channels=[response['pubsubChannelName']],
            uuid=client_id,
            subscribe_key=response['subscribeKey'],
            access_token=response['token'],
            renew_access_token_callback=renew_access_token
        )
        listener.on_message(handle_message)
        listener.start()
        return listener

def _random_string(num_chars: int) -> str:
    chars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
    return ''.join(random.choice(chars) for _ in range(num_chars))


from typing import Any, List, Protocol
import time

from pubnub.pnconfiguration import PNConfiguration
from pubnub.pubnub import PubNub
from pubnub.callbacks import SubscribeCallback
from pubnub.enums import PNStatusCategory

from pubnub.models.consumer.pubsub import PNMessageResult


class MessageCallback(Protocol):
    def __call__(self, *, channel: str, message: Any): ...

class RenewAccessTokenCallback(Protocol):
    def __call__(self) -> str: ...

class PubsubListener:
    def __init__(self,
        *,
        channels: List[str],
        uuid: str,
        access_token: str,
        renew_access_token_callback: RenewAccessTokenCallback,
        subscribe_key: str
    ) -> None:
        self._channels = channels
        self._uuid = uuid
        self._access_token = access_token
        self._renew_access_token_callback = renew_access_token_callback
        self._subscribe_key = subscribe_key
        self._message_callbacks: List[MessageCallback] = []
        self._ttl = None
        self._ttl_timestamp = None
    def on_message(self, callback: MessageCallback):
        self._message_callbacks.append(callback)
    def start(self):
        # token = self._get_access_token_callback(uuid=self._uuid, channel_names=self._channels)
        token = self._access_token

        pnconfig = PNConfiguration()

        pnconfig.subscribe_key = self._subscribe_key
        pnconfig.uuid = self._uuid
        self._pubnub = PubNub(pnconfig)
        self._pubnub.set_token(token)

        self._ttl = self._pubnub.parse_token(token)['ttl']
        self._ttl_timestamp = time.time()

        self._pubnub.add_listener(MySubscribeCallback(listener=self))

        self._pubnub.subscribe().channels(self._channels).execute()
    def wait(self, timeout_sec: float):
        self._check_token()
        time.sleep(timeout_sec)
    def stop(self):
        print('Unsubscribing from channels')
        self._pubnub.unsubscribe_all()
        time.sleep(1)
    def _check_token(self):
        if self._ttl is None:
            return
        elapsed_min = (time.time() - self._ttl_timestamp) / 60
        if elapsed_min + 1 > self._ttl:
            new_token = self._renew_access_token_callback()
            self._ttl = self._pubnub.parse_token(new_token)['ttl']
            self._ttl_timestamp = time.time()
            self._pubnub.set_token(new_token)

class MySubscribeCallback(SubscribeCallback):
    def __init__(self, listener: PubsubListener) -> None:
        self._listener = listener
        super().__init__()
    def presence(self, pubnub, presence):
        pass  # handle incoming presence data

    def status(self, pubnub, status):
        if status.category == PNStatusCategory.PNUnexpectedDisconnectCategory:
            pass  # This event happens when radio / connectivity is lost

        elif status.category == PNStatusCategory.PNConnectedCategory:
            # Connect event. You can do stuff like publish, and know you'll get it.
            # Or just use the connected event to confirm you are subscribed for
            # UI / internal notifications, etc
            pass
        elif status.category == PNStatusCategory.PNReconnectedCategory:
            pass
            # Happens as part of our regular operation. This event happens when
            # radio / connectivity is lost, then regained.
        elif status.category == PNStatusCategory.PNDecryptionErrorCategory:
            pass
            # Handle message decryption error. Probably client configured to
            # encrypt messages and on live data feed it received plain text.

    def message(self, pubnub, message: PNMessageResult):
        # Handle new message stored in message.message
        for cb in self._listener._message_callbacks:
            cb(channel=message.channel, message=message.message)


import os
import requests

from ._kachery_cloud_api_url import _kachery_cloud_api_url
from ._kacherycloud_request import _kacherycloud_request

def store_file(filename: str):
    # web3_storage_token = os.environ.get('HASHIO_WEB3_STORAGE_TOKEN', None)
    web3_storage_token = None
    # if not web3_storage_token and not hashio_api_url:
    #     raise Exception(f'Environment variable not set: HASHIO_WEB3_STORAGE_TOKEN or HASHIO_API_URL. See the hashio docs.')
    if web3_storage_token:
        url = 'https://api.web3.storage/upload'
        headers = {
            'Authorization': f'Bearer {web3_storage_token}'
        }
        with open(filename, 'rb') as f:
            resp = requests.post(url, data=f, headers=headers)
        if resp.status_code != 200:
            raise Exception(f'Error storing file ({resp.status_code}) {resp.reason}: {resp.text}')
        cid = resp.json()['cid']
        return f'ipfs://{cid}'
    elif _kachery_cloud_api_url:
        size = os.path.getsize(filename)
        payload = {
            'type': 'initiateIpfsUpload',
            'size': size
        }
        response = _kacherycloud_request(payload)
        signed_upload_url = response['signedUploadUrl']
        object_key = response['objectKey']
        with open(filename, 'rb') as f:
            resp_upload = requests.put(signed_upload_url, data=f)
            if resp_upload.status_code != 200:
                raise Exception(f'Error uploading file to bucket ({resp_upload.status_code}) {resp_upload.reason}: {resp_upload.text}')
        payload2 = {
            'type': 'finalizeIpfsUpload',
            'objectKey': object_key
        }
        response2 = _kacherycloud_request(payload2)
        cid = response2['cid']
        return f'ipfs://{cid}'
    else:
        raise Exception('Unexpected: no method found for uploading file')


try:
    import pubnub
except:
    # Let's not require pubnub by default - require user to install it manually if needed
    raise Exception('Pubnub not installed. To install run: pip install pubnub')

from .TaskBackend import TaskBackend
from .TaskClient import TaskClient
from .upload_task_result import upload_task_result, download_task_result